@@ -31,6 +31,7 @@
 #if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
 
 #include <X11/Xatom.h>
+#include <xf86.h>
 #include <xf86Xinput.h>
 #include <exevents.h>
 
@@ -38,6 +39,11 @@
 #include "synapticsstr.h"
 #include "synaptics-properties.h"
 
+#ifndef XATOM_FLOAT
+#define XATOM_FLOAT "FLOAT"
+#endif
+static Atom float_type;
+
 Atom prop_edges                 = 0;
 Atom prop_finger                = 0;
 Atom prop_tap_time              = 0;
@@ -70,6 +76,7 @@ Atom prop_palm                  = 0;
 Atom prop_palm_dim              = 0;
 Atom prop_coastspeed            = 0;
 Atom prop_pressuremotion        = 0;
+Atom prop_pressuremotion_factor = 0;
 Atom prop_grab                  = 0;
 
 static Atom
@@ -103,18 +110,21 @@ InitAtom(DeviceIntPtr dev, char *name, int format, int nvalues, int *values)
     atom = MakeAtom(name, strlen(name), TRUE);
     XIChangeDeviceProperty(dev, atom, XA_INTEGER, format,
                            PropModeReplace, nvalues,
-                           converted, FALSE, FALSE, FALSE);
+                           converted, FALSE);
+    XISetDevicePropertyDeletable(dev, atom, FALSE);
     return atom;
 }
 
-/**
- * Configure the property if the atom is valid.
- */
-static void
-ConfigureProperty(DeviceIntPtr dev, Atom atom, int nvalues, long* values)
+static Atom
+InitFloatAtom(DeviceIntPtr dev, char *name, int nvalues, float *values)
 {
-    if (atom)
-        XIConfigureDeviceProperty(dev, atom, FALSE, FALSE, FALSE, nvalues, values);
+    Atom atom;
+
+    atom = MakeAtom(name, strlen(name), TRUE);
+    XIChangeDeviceProperty(dev, atom, float_type, 32, PropModeReplace,
+                           nvalues, values, FALSE);
+    XISetDevicePropertyDeletable(dev, atom, FALSE);
+    return atom;
 }
 
 void
@@ -122,10 +132,20 @@ InitDeviceProperties(LocalDevicePtr local)
 {
     SynapticsPrivate *priv = (SynapticsPrivate *) local->private;
     SynapticsSHM *para = priv->synpara;
-
-    int i;
     int values[9]; /* we never have more than 9 values in an atom */
-    long valid_values[9];
+    float fvalues[4]; /* never have more than 4 float values */
+
+    float_type = XIGetKnownProperty(XATOM_FLOAT);
+    if (!float_type)
+    {
+        float_type = MakeAtom(XATOM_FLOAT, strlen(XATOM_FLOAT), TRUE);
+        if (!float_type)
+        {
+            xf86Msg(X_ERROR, "%s: Failed to init float atom. "
+                             "Disabling property support.\n", local->name);
+            return;
+        }
+    }
 
     values[0] = para->left_edge;
     values[1] = para->right_edge;
@@ -165,7 +185,11 @@ InitDeviceProperties(LocalDevicePtr local)
     values[1] = para->scroll_twofinger_horiz;
     prop_scrolltwofinger = InitAtom(local->dev, SYNAPTICS_PROP_SCROLL_TWOFINGER,8, 2, values);
 
-    /* FIXME: MISSING: speed is a float */
+    fvalues[0] = para->min_speed;
+    fvalues[1] = para->max_speed;
+    fvalues[2] = para->accl;
+    fvalues[3] = para->trackstick_speed;
+    prop_speed = InitFloatAtom(local->dev, SYNAPTICS_PROP_SPEED, 4, fvalues);
 
     values[0] = para->edge_motion_min_z;
     values[1] = para->edge_motion_max_z;
@@ -196,7 +220,10 @@ InitDeviceProperties(LocalDevicePtr local)
     prop_clickaction = InitAtom(local->dev, SYNAPTICS_PROP_CLICK_ACTION, 8, MAX_CLICK, values);
 
     prop_circscroll = InitAtom(local->dev, SYNAPTICS_PROP_CIRCULAR_SCROLLING, 8, 1, &para->circular_scrolling);
-    /* FIXME: missing: scroll_dist_circ is a float */
+
+    fvalues[0] = para->scroll_dist_circ;
+    prop_circscroll_dist = InitFloatAtom(local->dev, SYNAPTICS_PROP_CIRCULAR_SCROLLING_DIST, 1, fvalues);
+
     prop_circscroll_trigger = InitAtom(local->dev, SYNAPTICS_PROP_CIRCULAR_SCROLLING_TRIGGER, 8, 1, &para->circular_trigger);
     prop_circpad = InitAtom(local->dev, SYNAPTICS_PROP_CIRCULAR_PAD, 8, 1, &para->circular_pad);
     prop_palm = InitAtom(local->dev, SYNAPTICS_PROP_PALM_DETECT, 8, 1, &para->palm_detect);
@@ -206,61 +233,48 @@ InitDeviceProperties(LocalDevicePtr local)
 
     prop_palm_dim = InitAtom(local->dev, SYNAPTICS_PROP_PALM_DIMENSIONS, 32, 2, values);
 
-    /* FIXME: missing, coastspeed is a float */
+    fvalues[0] = para->coasting_speed;
+    prop_coastspeed = InitFloatAtom(local->dev, SYNAPTICS_PROP_COASTING_SPEED, 1, fvalues);
 
     values[0] = para->press_motion_min_z;
     values[1] = para->press_motion_max_z;
     prop_pressuremotion = InitAtom(local->dev, SYNAPTICS_PROP_PRESSURE_MOTION, 32, 2, values);
 
-    /* FIXME: missing, motion_min/max is a float */
+    fvalues[0] = para->press_motion_min_factor;
+    fvalues[1] = para->press_motion_max_factor;
+
+    prop_pressuremotion_factor = InitFloatAtom(local->dev, SYNAPTICS_PROP_PRESSURE_MOTION_FACTOR, 2, fvalues);
 
     prop_grab = InitAtom(local->dev, SYNAPTICS_PROP_GRAB, 8, 1, &para->grab_event_device);
 
-    /* Configure value ranges as appropriate */
-
-    /* boolean properties */
-    valid_values[0] = TRUE;
-    valid_values[0] = FALSE;
-    ConfigureProperty(local->dev, prop_tap_fast, 2, valid_values);
-    ConfigureProperty(local->dev, prop_scrolledge, 2, valid_values);
-    ConfigureProperty(local->dev, prop_scrolltwofinger, 2, valid_values);
-    ConfigureProperty(local->dev, prop_edgemotion_always, 2, valid_values);
-    ConfigureProperty(local->dev, prop_buttonscroll, 2, valid_values);
-    ConfigureProperty(local->dev, prop_buttonscroll_repeat, 2, valid_values);
-    ConfigureProperty(local->dev, prop_guestmouse, 2, valid_values);
-    ConfigureProperty(local->dev, prop_lockdrags, 2, valid_values);
-    ConfigureProperty(local->dev, prop_circscroll, 2, valid_values);
-    ConfigureProperty(local->dev, prop_circpad, 2, valid_values);
-    ConfigureProperty(local->dev, prop_palm, 2, valid_values);
-    ConfigureProperty(local->dev, prop_grab, 2, valid_values);
-    /* end of boolean props */
-
-    valid_values[0] = 0;
-    valid_values[1] = 1;
-    valid_values[2] = 2;
-    ConfigureProperty(local->dev, prop_off, 3, valid_values);
-
-    for (i = 0; i <= 8; i++)
-        valid_values[i] = i;
-    ConfigureProperty(local->dev, prop_circscroll_trigger, 8, valid_values);
 }
 
-Bool
-SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
+int
+SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
+            BOOL checkonly)
 {
     LocalDevicePtr local = (LocalDevicePtr) dev->public.devicePrivate;
     SynapticsPrivate *priv = (SynapticsPrivate *) local->private;
     SynapticsSHM *para = priv->synpara;
+    SynapticsSHM tmp;
+
+    /* If checkonly is set, no parameters may be changed. So just let the code
+     * change temporary variables and forget about it. */
+    if (checkonly)
+    {
+        tmp = *para;
+        para = &tmp;
+    }
 
     if (property == prop_edges)
     {
         INT32 *edges;
         if (prop->size != 4 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         edges = (INT32*)prop->data;
         if (edges[0] > edges[1] || edges[2] > edges[3])
-            return FALSE;
+            return BadValue;
 
         para->left_edge   = edges[0];
         para->right_edge  = edges[1];
@@ -271,11 +285,11 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     {
         INT32 *finger;
         if (prop->size != 3 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         finger = (INT32*)prop->data;
         if (finger[0] > finger[1])
-            return FALSE;
+            return BadValue;
 
         para->finger_low   = finger[0];
         para->finger_high  = finger[1];
@@ -284,14 +298,14 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     } else if (property == prop_tap_time)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->tap_time = *(INT32*)prop->data;
 
     } else if (property == prop_tap_move)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->tap_move = *(INT32*)prop->data;
     } else if (property == prop_tap_durations)
@@ -299,7 +313,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         INT32 *timeouts;
 
         if (prop->size != 3 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         timeouts = (INT32*)prop->data;
 
@@ -310,27 +324,27 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     } else if (property == prop_tap_fast)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->fast_taps = *(BOOL*)prop->data;
 
     } else if (property == prop_middle_timeout)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->emulate_mid_button_time = *(INT32*)prop->data;
     } else if (property == prop_twofinger_pressure)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->emulate_twofinger_z = *(INT32*)prop->data;
     } else if (property == prop_scrolldist)
     {
         INT32 *dist;
         if (prop->size != 2 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         dist = (INT32*)prop->data;
         para->scroll_dist_vert = dist[0];
@@ -339,7 +353,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     {
         CARD8 *edge;
         if (prop->size != 3 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         edge = (BOOL*)prop->data;
         para->scroll_edge_vert   = edge[0];
@@ -350,24 +364,34 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         CARD8 *twofinger;
 
         if (prop->size != 2 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         twofinger = (BOOL*)prop->data;
         para->scroll_twofinger_vert  = twofinger[0];
         para->scroll_twofinger_horiz = twofinger[1];
     } else if (property == prop_speed)
     {
-        /* XXX */
+        float *speed;
+
+        if (prop->size != 4 || prop->format != 32 || prop->type != float_type)
+            return BadMatch;
+
+        speed = (float*)prop->data;
+        para->min_speed = speed[0];
+        para->max_speed = speed[1];
+        para->accl = speed[2];
+        para->trackstick_speed = speed[3];
+
     } else if (property == prop_edgemotion_pressure)
     {
         CARD32 *pressure;
 
         if (prop->size != 2 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         pressure = (CARD32*)prop->data;
         if (pressure[0] > pressure[1])
-            return FALSE;
+            return BadValue;
 
         para->edge_motion_min_z = pressure[0];
         para->edge_motion_max_z = pressure[1];
@@ -377,11 +401,11 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         CARD32 *speed;
 
         if (prop->size != 2 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         speed = (CARD32*)prop->data;
         if (speed[0] > speed[1])
-            return FALSE;
+            return BadValue;
 
         para->edge_motion_min_speed = speed[0];
         para->edge_motion_max_speed = speed[1];
@@ -389,7 +413,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     } else if (property == prop_edgemotion_always)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->edge_motion_use_always = *(BOOL*)prop->data;
 
@@ -398,7 +422,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         BOOL *scroll;
 
         if (prop->size != 2 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         scroll = (BOOL*)prop->data;
         para->updown_button_scrolling    = scroll[0];
@@ -409,7 +433,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         BOOL *repeat;
 
         if (prop->size != 2 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         repeat = (BOOL*)prop->data;
         para->updown_button_repeat    = repeat[0];
@@ -417,7 +441,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     } else if (property == prop_buttonscroll_time)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->scroll_button_repeat = *(INT32*)prop->data;
 
@@ -425,30 +449,30 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     {
         CARD8 off;
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         off = *(CARD8*)prop->data;
 
         if (off > 2)
-            return FALSE;
+            return BadValue;
 
         para->touchpad_off = off;
     } else if (property == prop_guestmouse)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->guestmouse_off = *(BOOL*)prop->data;
     } else if (property == prop_lockdrags)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->locked_drags = *(BOOL*)prop->data;
     } else if (property == prop_lockdrags_time)
     {
         if (prop->size != 1 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->locked_drag_time = *(INT32*)prop->data;
     } else if (property == prop_tapaction)
@@ -457,7 +481,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         CARD8 *action;
 
         if (prop->size > MAX_TAP || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         action = (CARD8*)prop->data;
 
@@ -469,7 +493,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         CARD8 *action;
 
         if (prop->size > MAX_CLICK || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         action = (CARD8*)prop->data;
 
@@ -478,35 +502,41 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
     } else if (property == prop_circscroll)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->circular_scrolling = *(BOOL*)prop->data;
 
     } else if (property == prop_circscroll_dist)
     {
-        /* FIXME */
+        float circdist;
+
+        if (prop->size != 1 || prop->format != 32 || prop->type != float_type)
+            return BadMatch;
+
+        circdist = *(float*)prop->data;
+        para->scroll_dist_circ = circdist;
     } else if (property == prop_circscroll_trigger)
     {
         int trigger;
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         trigger = *(CARD8*)prop->data;
         if (trigger > 8)
-            return FALSE;
+            return BadValue;
 
         para->circular_trigger = trigger;
 
     } else if (property == prop_circpad)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->circular_pad = *(BOOL*)prop->data;
     } else if (property == prop_palm)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->palm_detect = *(BOOL*)prop->data;
     } else if (property == prop_palm_dim)
@@ -514,7 +544,7 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         INT32 *dim;
 
         if (prop->size != 2 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         dim = (INT32*)prop->data;
 
@@ -522,28 +552,35 @@ SetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop)
         para->palm_min_z     = dim[1];
     } else if (property == prop_coastspeed)
     {
-        /* FIXME */
+        float speed;
+
+        if (prop->size != 1 || prop->format != 32 || prop->type != float_type)
+            return BadMatch;
+
+        speed = *(float*)prop->data;
+        para->coasting_speed = speed;
+
     } else if (property == prop_pressuremotion)
     {
-        INT32 *press;
-        if (prop->size != 2 || prop->format != 32 || prop->type != XA_INTEGER)
-            return FALSE;
+        float *press;
+        if (prop->size != 2 || prop->format != 32 || prop->type != float_type)
+            return BadMatch;
 
-        press = (INT32*)prop->data;
+        press = (float*)prop->data;
         if (press[0] > press[1])
-            return FALSE;
+            return BadValue;
 
         para->press_motion_min_z = press[0];
         para->press_motion_max_z = press[1];
     } else if (property == prop_grab)
     {
         if (prop->size != 1 || prop->format != 8 || prop->type != XA_INTEGER)
-            return FALSE;
+            return BadMatch;
 
         para->grab_event_device = *(BOOL*)prop->data;
     }
 
-    return TRUE;
+    return Success;
 }
 
 #endif
